@@ -20,6 +20,7 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+import ahio
 import io
 import zipfile
 import hashlib
@@ -29,13 +30,17 @@ import logging
 import os.path
 import tempfile
 import scipy.io as sio
+from multiprocessing import Pipe
 
 from bson import json_util
 from enum import Enum
 
+from cheroot.wsgi import Server
+from cheroot.wsgi import PathInfoDispatcher
 from flask import Flask, request, send_file
 from moirai.database import DatabaseV1
 from moirai.hardware import Hardware
+from moirai import __version__
 
 
 class APIv1:
@@ -54,73 +59,106 @@ class APIv1:
         """
         Entry point for the class. Adds routes and starts listening.
         """
+
         @self.app.after_request
         def add_header(response):
             response.headers['Cache-Control'] = 'no-store'
             return response
-        self.app.add_url_rule('/',
-                              view_func=lambda: 'Moirai Control System\n')
-        self.app.add_url_rule('/login',
-                              view_func=self.login,
-                              methods=['POST'])
-        self.app.add_url_rule('/set-password',
-                              view_func=self.set_password,
-                              methods=['POST'])
-        self.app.add_url_rule('/hardware/drivers',
-                              view_func=self.hardware_drivers,
-                              methods=['GET'])
-        self.app.add_url_rule('/hardware/configuration',
-                              view_func=self.hardware_set_configuration,
-                              methods=['POST'])
-        self.app.add_url_rule('/hardware/configuration',
-                              view_func=self.hardware_get_configuration,
-                              methods=['GET'])
-        self.app.add_url_rule('/system_response/tests',
-                              view_func=self.system_response_get_tests,
-                              methods=['GET'])
-        self.app.add_url_rule('/system_response/tests',
-                              view_func=self.system_response_set_tests,
-                              methods=['POST'])
-        self.app.add_url_rule('/system_response/test/run',
-                              view_func=self.system_response_run,
-                              methods=['POST'])
-        self.app.add_url_rule('/system_response/test/stop',
-                              view_func=self.system_response_stop,
-                              methods=['GET'])
-        self.app.add_url_rule('/live_graph/tests',
-                              view_func=self.live_graph_list_tests,
-                              methods=['GET'])
-        self.app.add_url_rule('/live_graph/test',
-                              view_func=self.live_graph_get_test,
-                              methods=['POST'])
-        self.app.add_url_rule('/live_graph/test/remove',
-                              view_func=self.live_graph_remove_test,
-                              methods=['POST'])
-        self.app.add_url_rule('/live_graph/test/export',
-                              view_func=self.live_graph_export_mat,
-                              methods=['POST'])
-        self.app.add_url_rule('/controllers',
-                              view_func=self.controller_set,
-                              methods=['POST'])
-        self.app.add_url_rule('/controllers',
-                              view_func=self.controller_get,
-                              methods=['GET'])
-        self.app.add_url_rule('/controllers/run',
-                              view_func=self.controller_run,
-                              methods=['POST'])
-        self.app.add_url_rule('/controllers/stop',
-                              view_func=self.controller_stop,
-                              methods=['GET'])
-        self.app.add_url_rule('/dev/gen_dummy_tests',
-                              view_func=self.dev_gen_dummy_tests,
-                              methods=['GET'])
-        self.app.add_url_rule('/db/dump',
-                              view_func=self.dump_database,
-                              methods=['GET'])
-        self.app.add_url_rule('/db/restore',
-                              view_func=self.restore_database,
-                              methods=['POST'])
-        self.app.run(host="0.0.0.0", port=5000, threaded=True)
+
+        self.app.add_url_rule('/', view_func=lambda: 'Moirai Control System\n')
+        self.app.add_url_rule('/login', view_func=self.login, methods=['POST'])
+        self.app.add_url_rule(
+            '/version', view_func=self.version, methods=['GET'])
+        self.app.add_url_rule(
+            '/set-password', view_func=self.set_password, methods=['POST'])
+        self.app.add_url_rule(
+            '/last_error', view_func=self.last_error, methods=['GET'])
+        self.app.add_url_rule(
+            '/hardware/drivers',
+            view_func=self.hardware_drivers,
+            methods=['GET'])
+        self.app.add_url_rule(
+            '/hardware/configuration',
+            view_func=self.hardware_set_configuration,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/hardware/configuration',
+            view_func=self.hardware_get_configuration,
+            methods=['GET'])
+        self.app.add_url_rule(
+            '/system_response/tests',
+            view_func=self.system_response_get_tests,
+            methods=['GET'])
+        self.app.add_url_rule(
+            '/system_response/tests',
+            view_func=self.system_response_set_tests,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/system_response/test/run',
+            view_func=self.system_response_run,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/system_response/test/stop',
+            view_func=self.system_response_stop,
+            methods=['GET'])
+        self.app.add_url_rule(
+            '/live_graph/tests',
+            view_func=self.live_graph_list_tests,
+            methods=['GET'])
+        self.app.add_url_rule(
+            '/live_graph/test',
+            view_func=self.live_graph_get_test,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/live_graph/test/remove',
+            view_func=self.live_graph_remove_test,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/live_graph/test/export',
+            view_func=self.live_graph_export_mat,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/controllers', view_func=self.controller_set, methods=['POST'])
+        self.app.add_url_rule(
+            '/controllers', view_func=self.controller_get, methods=['GET'])
+        self.app.add_url_rule(
+            '/controllers/run',
+            view_func=self.controller_run,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/controllers/export',
+            view_func=self.controller_export,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/controllers/import',
+            view_func=self.controller_import,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/controllers/stop',
+            view_func=self.controller_stop,
+            methods=['GET'])
+        self.app.add_url_rule(
+            '/db/dump', view_func=self.dump_database, methods=['GET'])
+        self.app.add_url_rule(
+            '/db/restore', view_func=self.restore_database, methods=['POST'])
+        self.app.add_url_rule(
+            '/simulation/run',
+            view_func=self.model_simulation_run,
+            methods=['POST'])
+        self.app.add_url_rule(
+            '/pid/run', view_func=self.pid_run, methods=['POST'])
+        self.app.add_url_rule(
+            '/free/run', view_func=self.free_run, methods=['POST'])
+
+        d = PathInfoDispatcher({'/': self.app})
+        self.server = Server(('0.0.0.0', 5000), d)
+        self.server.start()
+
+    def stop(self):
+        self.server.stop()
+
+    def version(self):
+        return json.dumps({'version': __version__})
 
     def verify_token(self):
         """
@@ -130,8 +168,7 @@ class APIv1:
             authorization = request.headers.get('Authorization')
             token = authorization.split(' ')[-1]
             return self.database.verify_token(token)
-        except Exception as e:
-            print(e)
+        except Exception:  # noqa: E722 pylint: disable=E722
             return False
 
     def login(self):
@@ -189,6 +226,20 @@ class APIv1:
         self.database.set_setting('password', password)
         return '{}'
 
+    def last_error(self):
+        """
+        Returns the last error in the database.
+
+        @returns:
+            {
+                message: string
+            }
+        """
+        if not self.verify_token():
+            return '{}', 403
+        error = self.database.get_setting('test_error')
+        return json.dumps({'message': error})
+
     def hardware_drivers(self):
         """
         Returns a JSON object with all the available drivers, their setups, if
@@ -227,10 +278,14 @@ class APIv1:
             return '{}', 403
         drivers = self.hardware.list_drivers()
         drivers = [{
-            'name': driver,
-            'has_setup': self.hardware.driver_has_setup(driver),
-            'setup_arguments': self.hardware.driver_setup_arguments(driver),
-            'ports': self.__ports_for_driver(driver)
+            'name':
+            driver,
+            'has_setup':
+            self.hardware.driver_has_setup(driver),
+            'setup_arguments':
+            self.hardware.driver_setup_arguments(driver),
+            'ports':
+            self.__ports_for_driver(driver)
         } for driver in drivers]
         return json.dumps(drivers)
 
@@ -340,7 +395,7 @@ class APIv1:
                 name: string
                 type: string
                 inputs: string[]
-                output: string
+                output: string[]
                 points: [{
                     x: number
                     y: number
@@ -376,7 +431,7 @@ class APIv1:
             name: string
             type: string
             inputs: string[]
-            output: string
+            output: string[]
             points: [{
                 x: number
                 y: number
@@ -471,15 +526,17 @@ class APIv1:
             return '{}', 403
         running_test = self.database.get_setting('current_test')
         tests = self.database.list_test_data()
-
         if len(tests) == 0:
             return '[]'
 
         last_date = max([test['date'] for test in tests])
         tests = [{
-            'name': t['name'],
-            'date': t['date'].isoformat(),
-            'running': t['name'] == running_test and t['date'] == last_date
+            'name':
+            t['name'],
+            'date':
+            t['date'].isoformat(),
+            'running':
+            t['name'] == running_test and t['date'] == last_date
         } for t in tests]
 
         return json.dumps(tests)
@@ -546,13 +603,13 @@ class APIv1:
         ts = request.json
         if isinstance(ts, list):
             ts = [{
-                'test': t['test'],
-                'start_time': dateutil.parser.parse(t['start_time'])
+                'name': t['test'],
+                'date': dateutil.parser.parse(t['start_time'])
             } for t in ts]
         else:
             ts = {
-                'test': ts['test'],
-                'start_time': dateutil.parser.parse(ts['start_time'])
+                'name': ts['test'],
+                'date': dateutil.parser.parse(ts['start_time'])
             }
 
         self.database.remove_test(ts)
@@ -588,7 +645,7 @@ class APIv1:
         ks = list(v.keys())
         ds = self.database.get_filtered_test_data(test, start_time, ks)
         ts = ds[0]['time']
-        ds = {d['sensor']: d['values'] for d in ds}
+        ds = {v.get(d['sensor'], d['sensor']): d['values'] for d in ds}
         ds['t'] = ts
 
         directory = tempfile.gettempdir()
@@ -683,6 +740,76 @@ class APIv1:
         self.ph.send_command("hardware", "run_controller", controller)
         return '{}'
 
+    def controller_export(self):
+        """
+        Exports the given controllers. It must be a POST request with the
+        following body:
+
+        {
+            controllers: number[]
+        }
+
+        @returns: On success, HTTP 200 Ok and body:
+
+            file-contents
+
+            On failure, HTTP 403 Unauthorized and body:
+
+            {}
+        """
+        if not self.verify_token():
+            return '{}', 403
+
+        controllers = request.json['controllers']
+        cs = self.database.get_setting('controllers') or []
+        cs = [c for c in cs if c['id'] in controllers]
+        for c in cs:
+            c.pop('id', None)
+        jsondata = json.dumps(cs)
+
+        zbuffer = io.BytesIO()
+        with zipfile.ZipFile(zbuffer, "a", zipfile.ZIP_LZMA) as zip_file:
+            zip_file.writestr('dump', jsondata)
+        zbuffer.seek(0, 0)
+        return send_file(
+            zbuffer,
+            as_attachment=True,
+            attachment_filename='dump.zip',
+            mimetype="application/octet-stream")
+
+    def controller_import(self):
+        """
+        Restore controllers. It must be a POST request with following body:
+
+        file-contents
+
+        @returns:
+            On success, HTTP 200 Ok and body:
+
+            {}
+
+            On failure, HTTP 403 Unauthorized and body:
+
+            {}
+        """
+        if not self.verify_token():
+            return '{}', 403
+        file = tempfile.TemporaryFile()
+        file.write(request.files['file'].read())
+        file.seek(0)
+        with zipfile.ZipFile(file, "r", zipfile.ZIP_LZMA) as zip_file:
+            jsondata = zip_file.read('dump')
+            ns = json.loads(jsondata, object_hook=json_util.object_hook)
+            del jsondata
+            cs = self.database.get_setting('controllers') or []
+            lid = max([c['id'] for c in cs] or [0])
+            for n in ns:
+                lid += 1
+                n['id'] = lid
+                cs.append(n)
+            self.database.set_setting('controllers', cs)
+        return '{}'
+
     def controller_stop(self):
         """
         Stops the running controller. It must be a GET request.
@@ -722,24 +849,26 @@ class APIv1:
         if not self.verify_token():
             return '{}', 403
 
-        settings, test_sensor_values = self.database.dump_database()
-        data = {
-            'settings': settings,
-            'test_sensor_values': test_sensor_values
-        }
+        settings, graphs = self.database.dump_database()
+        data = {'settings': settings, 'graphs': graphs}
         jsondata = json.dumps(data, default=json_util.default)
 
-        del data, settings, test_sensor_values  # release memory
+        del data, settings, graphs  # release memory
 
         zbuffer = io.BytesIO()
         with zipfile.ZipFile(zbuffer, "a", zipfile.ZIP_LZMA) as zip_file:
             zip_file.writestr('dump', jsondata)
         zbuffer.seek(0, 0)
-        return send_file(zbuffer, as_attachment=True, attachment_filename='dump.zip', mimetype="application/octet-stream")
+        return send_file(
+            zbuffer,
+            as_attachment=True,
+            attachment_filename='dump.zip',
+            mimetype="application/octet-stream")
 
     def restore_database(self):
         """
-        Restore the database collections. It must be a POST request with following body:
+        Restore the database collections. It must be a POST request with
+        following body:
 
         [{
             settings: []
@@ -758,12 +887,17 @@ class APIv1:
         if not self.verify_token():
             return '{}', 403
         db = self.database
-        file = request.files['file']
+        file = tempfile.TemporaryFile()
+        file.write(request.files['file'].read())
+        file.seek(0)
         with zipfile.ZipFile(file, "r", zipfile.ZIP_LZMA) as zip_file:
             jsondata = zip_file.read('dump')
             d = json.loads(jsondata, object_hook=json_util.object_hook)
             del jsondata
-            db.restore_database(d['settings'], d['test_sensor_values'])
+            if 'test_sensor_values' in d:
+                db.restore_database_v1(d['settings'], d['test_sensor_values'])
+            else:
+                db.restore_database_v2(d['settings'], d['graphs'])
         return '{}'
 
     def __ports_for_driver(self, driver):
@@ -776,8 +910,8 @@ class APIv1:
 
     def __encode_port(self, port):
         """
-        If the port is a Enum, use its value. If it's a value that can be dumped
-        to JSON, use it, if not, raise an exception.
+        If the port is a Enum, use its value. If it's a value that can be
+        dumped to JSON, use it, if not, raise an exception.
         """
         port_id = port['id']
         if isinstance(port['id'], Enum):
@@ -785,7 +919,7 @@ class APIv1:
         else:
             try:
                 port_id = json.dumps(port['id'])
-            except:
+            except BaseException:
                 pass
         port['id'] = port_id
         return port
@@ -805,26 +939,93 @@ class APIv1:
                 port['id'] = p['id'].__class__(port['id'])
         return port
 
-    def dev_gen_dummy_tests(self):
-        import datetime
-        start_time = datetime.datetime.utcnow()
-        ts = list(range(600))
+    def model_simulation_run(self):
+        """
+        Simulates the model. It must be a POST request with the following body:
 
-        xs = ts
-        ys = [x ** 2 for x in xs]
-        zs = [2 * x for x in xs]
+        {
+            model: string
+            x0: string
+            u: string
+            duration: string
+        }
 
-        vss = [xs, ys, zs]
-        ns = ['x', 'y', 'z']
+        @returns: On success, HTTP 200 Ok and body:
 
-        for n, vs in zip(ns, vss):
-            for x, t in zip(vs, ts):
-                self.database.save_test_sensor_value(
-                    'Dummy',
-                    n,
-                    x,
-                    t,
-                    start_time
-                )
+            {}
 
-        return '[]'
+            On failure, HTTP 403 Unauthorized and body:
+
+            {}
+        """
+        if not self.verify_token():
+            return '{}', 403
+
+        args = request.json
+        p1, p2 = Pipe()
+        self.ph.send_command("hardware", "run_simulation", (args, p1))
+        return p2.recv()
+
+    def pid_run(self):
+        """
+        Runs the PID controller for controlled-variable y (port name expected)
+        and control-variable u (port name expected).
+
+        {
+            Kp: number
+            Ki: number
+            Kd: number
+            dt: number
+            u: string
+            y: string
+            r: string
+            umax: number
+            umin: number
+            fixedOutputs: {
+                            alias: string
+                            value: number
+                          }[]
+        }
+
+        @returns: On success, HTTP 200 Ok and body:
+
+            {}
+
+            On failure, HTTP 403 Unauthorized and body:
+
+            {}
+        """
+        if not self.verify_token():
+            return '{}', 403
+
+        args = request.json
+        self.ph.send_command("hardware", "run_pid", args)
+        return '{}'
+
+    def free_run(self):
+        """
+        Controls the plant outputs freely.
+
+        {
+            dt: number
+            inputs: string[]
+            outputs: {
+                        alias: string
+                        value: number
+                     }[]
+        }
+
+        @returns: On success, HTTP 200 Ok and body:
+
+            {}
+
+            On failure, HTTP 403 Unauthorized and body:
+
+            {}
+        """
+        if not self.verify_token():
+            return '{}', 403
+
+        args = request.json
+        self.ph.send_command("hardware", "run_free", args)
+        return '{}'
